#include "Utils.h"
#include "ProbabilityDistributions.h"
#include "RNG.h"
#include "BivariateGMM.h"

namespace bcm {

BivariateGMM::BivariateGMM()
	: num_components(0)
{
}

BivariateGMM::~BivariateGMM()
{
}

bool BivariateGMM::Initialize(boost::property_tree::ptree pt)
{
	try {
		num_components = pt.get<size_t>("<xmlattr>.num_components");
		if (num_components == 0) {
			LOGERROR("GMM must have at least 1 component");
			return false;
		}

		std::string proportions_str = pt.get<std::string>("<xmlattr>.proportions");
		std::string mu1_str = pt.get<std::string>("<xmlattr>.mu1");
		std::string mu2_str = pt.get<std::string>("<xmlattr>.mu2");
		std::string sigma1_str = pt.get<std::string>("<xmlattr>.sigma1");
		std::string sigma2_str = pt.get<std::string>("<xmlattr>.sigma2");
		std::string correlation_str = pt.get<std::string>("<xmlattr>.correlation");

		std::vector<std::string> proportions_v;
		std::vector<std::string> mu1_v;
		std::vector<std::string> mu2_v;
		std::vector<std::string> sigma1_v;
		std::vector<std::string> sigma2_v;
		std::vector<std::string> correlation_v;

		tokenize(proportions_str, proportions_v, ",");
		tokenize(mu1_str, mu1_v, ",");
		tokenize(mu2_str, mu2_v, ",");
		tokenize(sigma1_str, sigma1_v, ",");
		tokenize(sigma2_str, sigma2_v, ",");
		tokenize(correlation_str, correlation_v, ",");

		if (proportions_v.size() != num_components) {
			LOGERROR("Inconsistent number of elements in proportions, should have %u elements", num_components);
			return false;
		}
		if (mu1_v.size() != num_components) {
			LOGERROR("Inconsistent number of elements in mu1, should have %u elements", num_components);
			return false;
		}
		if (mu2_v.size() != num_components) {
			LOGERROR("Inconsistent number of elements in mu2, should have %u elements", num_components);
			return false;
		}
		if (sigma1_v.size() != num_components) {
			LOGERROR("Inconsistent number of elements in sigma1, should have %u elements", num_components);
			return false;
		}
		if (sigma2_v.size() != num_components) {
			LOGERROR("Inconsistent number of elements in sigma2, should have %u elements", num_components);
			return false;
		}
		if (correlation_v.size() != num_components) {
			LOGERROR("Inconsistent number of elements in correlation, should have %u elements", num_components);
			return false;
		}

		proportions.resize(num_components);
		mu1.resize(num_components);
		mu2.resize(num_components);
		sigma1.resize(num_components);
		sigma2.resize(num_components);
		correlation.resize(num_components);

		for (size_t i = 0; i < num_components; i++) {
			proportions(i) = atof(proportions_v[i].c_str());
			mu1(i) = atof(mu1_v[i].c_str());
			mu2(i) = atof(mu2_v[i].c_str());
			sigma1(i) = atof(sigma1_v[i].c_str());
			sigma2(i) = atof(sigma2_v[i].c_str());
			correlation(i) = atof(correlation_v[i].c_str());
		}
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing UnivariateMarginal: %s", e.what());
		return false;
	}

	return true;
}

bool BivariateGMM::Sample(Real& x, Real& y, RNG* rng) const
{
	// Sample a component
	size_t c = num_components - 1;
	Real p = rng->GetReal();
	Real t = 0;
	for (size_t i = 0; i < num_components - 1; i++) {
		t += proportions[i];
		if (p < t) {
			c = i;
			break;
		}
	}

	// Sample from that components
	Real nx = rng->GetNormal(0.0, 1.0);
	Real ny = rng->GetNormal(0.0, 1.0);
	Real rho = correlation[c];

	x = mu1[c] + sigma1[c] * nx;
	y = mu2[c] + sigma2[c] * (rho * nx + sqrt(1.0 - rho * rho) * ny);

	return true;
}

bool BivariateGMM::EvaluatePDF(Real& p, const Real x, const Real y) const
{
	p = 0.0;
	for (size_t i = 0; i < num_components; i++) {
		Real rho = correlation[i];
		Real dx = x - mu1[i];
		Real dy = y - mu2[i];

		Real e = (-1.0 / (2.0 * (1.0 - rho * rho))) * 
			(dx*dx/(sigma1[i]*sigma1[i]) +
			 dy*dy/(sigma2[i]*sigma2[i]) -
			 dx*dy/(sigma1[i]*sigma2[i]) * 2.0 * rho);
		Real c = 1.0 / (2.0 * M_PI * sigma1[i] * sigma2[i] * sqrt(1.0 - rho * rho));
		Real thisp = c * exp(e);

		p += proportions[i] * thisp;
	}
	return true;
}

bool BivariateGMM::EvaluateLogPDF(Real& logp, const Real x, const Real y) const
{
	Real p;
	if (!EvaluatePDF(p, x, y)) {
		return false;
	}
	logp = log(p);
	return true;
}

}
