#include "Utils.h"
#include "OutputValueHandlerFileDumper.h"
#include "VariableSet.h"

namespace bcm {

OutputValueHandlerFileDumper::OutputValueHandlerFileDumper(size_t buffer_size)
	: num_values(0)
	, Initialized(false)
	, output_buffer_ix(0)
{
	output_buffer.resize(buffer_size);
}

OutputValueHandlerFileDumper::~OutputValueHandlerFileDumper()
{
	if (Initialized) {
		Flush();
	}
}

void OutputValueHandlerFileDumper::SetFile(const std::string& filename)
{
	this->filename = filename;
	
	std::string extension;
	size_t extension_pos = filename.find_last_of(".");
	if (extension_pos != std::string::npos) {
		extension = filename.substr(extension_pos + 1);
		std::transform(extension.begin(), extension.end(), extension.begin(), ::tolower);
		if (extension == "tsv" || extension == "txt") {
			binary = false;
		} else {
			binary = true;
		}
	} else {
		binary = true;
	}
}

void OutputValueHandlerFileDumper::SetNumValues(size_t n)
{
	num_values = n;
	for (size_t i = 0; i < output_buffer.size(); i++) {
		output_buffer[i] = VectorReal::Zero(n);
	}
}

bool OutputValueHandlerFileDumper::Initialize()
{
	if (filename.empty()) {
		LOGERROR("Output filename has not been specified");
		return false;
	}

	if (Initialized) {
		Flush();
		Initialized = false;
	}

	FILE* file;
	if (binary) {
		file = fopen(filename.c_str(), "wb");
	} else {
		file = fopen(filename.c_str(), "w");
	}

	if (!file) {
		LOGERROR("Failed to open output file \"%s\"", filename.c_str());
		return false;
	} else {
		if (!binary) {
			for (size_t i = 0; i < num_values; i++) {
				if (i == num_values - 1) {
					fprintf(file, "output_%u\n", (unsigned int)i);
				} else {
					fprintf(file, "output_%u\t", (unsigned int)i);
				}
			}
		}
		fclose(file);
	}

	output_buffer_ix = 0;
	Initialized = true;
	return true;
}

void OutputValueHandlerFileDumper::Handle(const Real* values, void*)
{
	ASSERT(Initialized);

	VectorReal& sample = output_buffer[output_buffer_ix];
	for (size_t i = 0; i < num_values; i++) {
		sample(i) = values[i];
	}

	output_buffer_ix++;
	if (output_buffer_ix == output_buffer.size()) {
		Flush();
	}
}

boost::function<void (const Real* values, void* user)> OutputValueHandlerFileDumper::GetFunctionPointer()
{
	ASSERT(Initialized);
	return boost::bind(&bcm::OutputValueHandlerFileDumper::Handle, this, _1, _2);
}

void OutputValueHandlerFileDumper::Flush()
{
	FILE* file;
	if (binary) {
		file = fopen(filename.c_str(), "ab");
	} else {
		file = fopen(filename.c_str(), "a");
	}
	if (file) {
		for (size_t j = 0; j < output_buffer_ix; j++) {
			const VectorReal& sample = output_buffer[j];
				
			if (binary) {
				float x;
				for (size_t i = 0; i < num_values; i++) {
					x = (float)sample(i);
					fwrite(&x, 4, 1, file);
				}
			} else {
				for (size_t i = 0; i < num_values; i++) {
					if (i == num_values - 1) {
						fprintf(file, "%.6g\n", sample(i));
					} else {
						fprintf(file, "%.6g\t", sample(i));
					}
				}
			}
		}
		fclose(file);
	} else {
		LOGERROR("Failed to open output file \"%s\"", filename.c_str());
	}
	output_buffer_ix = 0;
}

}
