#include "Utils.h"
#include "RNG_SFMT.h"
#include "SamplerSMC.h"
#include "SMCSamplingStrategy.h"
#include "VariableSet.h"

namespace bcm {

SMCSamplingStrategy::SMCSamplingStrategy(SamplerSMC* sampler, TaskManager* task_manager)
	: sampler(sampler)
	, prior(sampler->prior)
	, task_manager(task_manager)
	, num_variables(sampler->varset->GetNumVariables())
	, population_size(sampler->population_size)
	, num_threads(sampler->num_threads)
	, previous_population(NULL)
	, current_temperature(0.0)
	, previous_temperature(0.0)
{
}

SMCSamplingStrategy::~SMCSamplingStrategy()
{
}

bool SMCSamplingStrategy::UpdateStrategyIteration(bool force_update, RNG* rng)
{
	this->previous_population = sampler->previous_population;
	current_temperature = sampler->current_temperature;
	previous_temperature = sampler->previous_temperature;
	if (!UpdateStrategyIterationImpl(force_update, rng)) {
		return false;
	}
	return UpdateStrategyCurrent(rng);
}

bool SMCSamplingStrategy::UpdateStrategyCurrent(RNG* rng)
{
	return UpdateStrategyCurrentImpl(rng);
}

bool SMCSamplingStrategy::UpdateParticle(SamplerSMC::Particle& particle, size_t ix, RNG* rng, size_t threadix)
{
	return UpdateParticleImpl(particle, ix, rng, threadix);
}

bool SMCSamplingStrategy::CalculateParticleWeight(SamplerSMC::Particle& particle, size_t ix, size_t threadix)
{
	return CalculateParticleWeightImpl(particle, ix, threadix);
}

bool SMCSamplingStrategy::EvaluatePriorLikelihood(size_t threadix, const Real* values, Real& lprior, Real& llh)
{
	return sampler->EvaluatePriorLikelihood(threadix, values, lprior, llh);
}

bool SMCSamplingStrategy::GetOutputValues(size_t threadix, Real* values)
{
	return sampler->GetOutputValues(threadix, values);
}

VectorReal SMCSamplingStrategy::CalculatePreviousPopulationMean()
{
	VectorReal empirical_mean = VectorReal::Zero(num_variables);
	Real acc_weight = 0.0;
	for (size_t k = 0; k < previous_population->size(); k++) {
		const VectorReal& s = (*previous_population)[k].var_values;
		const Real w = (*previous_population)[k].normweight * previous_population->size();
		if (w > 0.0) {
			Real one_over_kp1 = w / (acc_weight + w);
			for (size_t i = 0; i < num_variables; i++) {
				empirical_mean(i) += (s(i) - empirical_mean(i)) * one_over_kp1;
			}
			acc_weight += w;
		}
	}
	return empirical_mean;
}

VectorReal SMCSamplingStrategy::CalculatePreviousPopulationVariance(const VectorReal& mean)
{
	VectorReal empirical_variance = VectorReal::Zero(num_variables);
	Real acc_weight = 0.0;
	for (size_t k = 0; k < previous_population->size(); k++) {
		const VectorReal& s = (*previous_population)[k].var_values;
		const Real w = (*previous_population)[k].normweight * previous_population->size();
		if (w > 0.0) {
			Real one_over_kp1 = w / (acc_weight + w);
			for (size_t i = 0; i < num_variables; i++) {
				Real d = s(i) - mean(i);
				empirical_variance(i) += (d * d - empirical_variance(i)) * one_over_kp1;
			}
			acc_weight += w;
		}
	}
	Real n = (Real)previous_population->size();
	empirical_variance *= n / (n - 1.0);
	return empirical_variance;
}

MatrixReal SMCSamplingStrategy::CalculatePreviousPopulationCovariance(const VectorReal& mean)
{
	MatrixReal empirical_covariance = MatrixReal::Zero(num_variables, num_variables);
	Real acc_weight = 0.0;
	for (size_t k = 0; k < previous_population->size(); k++) {
		const VectorReal& s = (*previous_population)[k].var_values;
		const Real w = (*previous_population)[k].normweight * previous_population->size();
		if (w > 0.0) {
			Real one_over_kp1 = w / (acc_weight + w);
			for (size_t i = 0; i < num_variables; i++) {
				for (size_t j = i; j < num_variables; j++) {
					Real di = s(i) - mean(i);
					Real dj = s(j) - mean(j);
					empirical_covariance(i,j) += (di * dj - empirical_covariance(i,j)) * one_over_kp1;
				}
			}
			acc_weight += w;
		}
	}
	Real n = (Real)previous_population->size();
	empirical_covariance *= n / (n - 1.0);
	return empirical_covariance;
}

}
