#pragma once

#include <boost/program_options.hpp>
#include "Serialization.h"
#include "TaskManager.h"

namespace bcm {
	
class Likelihood;
class Prior;
class ProgressIndicator;
class VariableSet;

class Sampler
{
public:
	Sampler(size_t num_threads);
	virtual ~Sampler();

	void SetLatentVariableSet(VariableSet* varset);
	void SetReentrantPrior(Prior* prior);
	void SetReentrantLikelihood(Likelihood* likelihood);
	bool SetNonreentrantPrior(size_t i, Prior* prior);
	bool SetNonreentrantLikelihood(size_t i, Likelihood* likelihood);
	void SetOutputPath(const std::string path);

	virtual size_t GetNumSamples() = 0;

	typedef boost::function<void (const Real* var_values, Real lposterior, Real llh, Real lprior, Real weight, void* user)> TPosteriorSampleHandler;
	typedef boost::function<void (const Real* output_values, void* user)> TOutputValuesHandler;
	typedef boost::function<void (Real evidence, void* user)> TEvidenceHandler;

	void SetPosteriorSampleHandler(TPosteriorSampleHandler fp, void* user);
	void SetOutputValuesHandler(TOutputValuesHandler fp, void* user);
	void SetEvidenceHandler(TEvidenceHandler fp, void* user);
	void SetProgressIndicator(ProgressIndicator* pi);

	bool Parse(const boost::program_options::variables_map& vm);
	bool Initialize();
	void Uninitialize();
	bool Run(Real* levidence = NULL, Real* levidence_sd = NULL);

    friend class boost::serialization::access;
    template<class Archive> void serialize(Archive& ar, const unsigned int version);

protected:
	enum ERunResult {
		RUN_FINISHED,
		RUN_TIMED_END,
		RUN_ERROR,
	};

	virtual bool ParseImpl(const boost::program_options::variables_map& vm) { return true; }
	virtual bool InitializeImpl() { return true; }
	virtual void UninitializeImpl() {}
	virtual ERunResult RunImpl() = 0;

	bool EvaluatePrior(size_t threadix, const Real* values, Real& lprior);
	bool EvaluateLikelihood(size_t threadix, const Real* values, Real& llh);
	bool EvaluatePriorLikelihood(size_t threadix, const Real* values, Real& lprior, Real& llh);
	bool GetOutputValues(size_t threadix, Real* values);

	bool SaveCurrentState();
	bool LoadCurrentState();

	// Settings
	bool have_reentrant_prior;
	bool have_reentrant_likelihood;

	Prior* prior;
	Likelihood* likelihood;

	std::vector<Prior*> nonreentrant_priors;
	std::vector<Likelihood*> nonreentrant_likelihoods;

	VariableSet* varset;
	unsigned long rngseed;
	size_t num_variables;
	size_t num_output_values;
	size_t num_threads;

	TPosteriorSampleHandler posterior_sample_handler;
	TOutputValuesHandler output_values_handler;
	TEvidenceHandler evidence_handler;
	void* posterior_sample_handler_user;
	void* output_values_handler_user;
	void* evidence_handler_user;

	std::string output_path;
	ProgressIndicator* progress;
	
	size_t runtime_seconds;						///< Run inference for this many seconds, dumping the current state at the end
	std::string	multiple_start_state_file;		///< Folder in which end result is dumped (and starting point is read in).
	bool multistart_continue;

	// Run-time variables
	bool initialized;
	size_t num_likelihood_evaluations;
	std::vector<size_t> num_parallel_likelihood_evaluations;
	Real evidence;
	Real evidence_sd;

	TaskManager* task_manager;
};

template<class Archive> void Sampler::serialize(Archive& ar, const unsigned int version)
{
	ar & evidence;
	ar & evidence_sd;
}

}
