#include "Utils.h"
#include "Likelihood.h"
#include "Prior.h"
#include "PriorBivariateMarginalGMM.h"
#include "PriorMarginalGMM.h"
#include "ProgressIndicator.h"
#include "RNG_SFMT.h"
#include "SamplerIS.h"
#include "VariableSet.h"

namespace bcm {

SamplerIS::SamplerIS(size_t num_threads)
	: Sampler(num_threads)
	, num_samples(1000)
	, rng(NULL)
	, proposal(NULL)
{
}

SamplerIS::~SamplerIS()
{
	safe_delete(rng);
	safe_delete(proposal);
}

bool SamplerIS::ParseImpl(const boost::program_options::variables_map& vm)
{
	try {
		num_samples = vm["sampler.sampling_period"].as<size_t>();
		proposal_file = vm["sampler.is_proposal_gmm"].as<std::string>();
	} catch (boost::program_options::error& e) {
		LOGERROR("Error parsing sampler: %s", e.what());
		return false;
	}

	return true;
}

bool SamplerIS::InitializeImpl()
{
	rng = new RNG_SFMT(rngseed);

	PriorMarginalGMM* prop = new PriorMarginalGMM(varset);
	//PriorBivariateMarginalGMM* prop = new PriorBivariateMarginalGMM(varset);
	if (!prop->LoadFromXML(proposal_file)) {
		return false;
	}
	proposal = prop;

	return true;
}

Sampler::ERunResult SamplerIS::RunImpl()
{
	VectorReal sample(num_variables);
	VectorReal output(num_output_values);

	for (size_t si = 0; si < num_samples; si++) {
		if (si % 10 == 0) {
			progress->ReportProgress(si / (Real)num_samples);
		}

		if (!proposal->Sample(sample.data(), rng)) {
			return RUN_ERROR;
		}

		Real lproposal;
		if (!proposal->EvaluateLogPDF(0, sample.data(), lproposal)) {
			return RUN_ERROR;
		}

		Real lprior, llh;
		if (!EvaluatePriorLikelihood(0, sample.data(), lprior, llh)) {
			return RUN_ERROR;
		}

		Real lposterior = lprior + llh;
		Real lweight = lposterior - lproposal;

		if (posterior_sample_handler) {
			posterior_sample_handler(sample.data(), lposterior, llh, lprior, exp(lweight), posterior_sample_handler_user);
		}
		if (output_values_handler && num_output_values > 0) {
			if (!GetOutputValues(0, output.data())) {
				return RUN_ERROR;
			}
			output_values_handler(output.data(), output_values_handler_user);
		}
	}

	progress->ReportProgress(1.0);

	return RUN_FINISHED;
}

}
