#pragma once

#include "RNG.h"
#include "Sample.h"
#include "SamplePopulation.h"
#include "Sampler.h"

namespace bcm {

class NestedSamplingStrategy;

class SamplerNested : public Sampler
{
public:
	SamplerNested(size_t num_threads);
	virtual ~SamplerNested();

	virtual size_t GetNumSamples() { return num_posterior_samples; }
	
private:
	virtual bool ParseImpl(const boost::program_options::variables_map& vm);
	virtual bool InitializeImpl();
	virtual void UninitializeImpl();
	virtual ERunResult RunImpl();

	bool SampleInitialPopulation();
	void AddFinalContribution(size_t steps_taken, Real lweight_factor);
	void CalculateEvidenceErrorEstimate();
	Real CalculateAverageEvidence();
	Real CalculateAverageInformation();
	void UpdateProgress(Real p);

	void GetNewSampleAsync(void* user, size_t threadix);
	void EmitPosteriorSample(const Sample* sample, Real lposterior_weight);

	enum ETerminationCriterion {
		SC_FixedStepNumber,				//< Run for a fixed number of steps, specified by termination_num_steps
		SC_NtimesHSteps,				//< Run until number of steps >= factor * population_size * information (original criterion from Skilling's 2004 paper is factor=2)
		SC_RemainingContribution,		//< Run until the current population of active samples can contribute only termination_remaining_contribution_threshold to the evidence (criterion from Feroz & Hobson's MultiNest sampler)
	};

	// Settings
	size_t population_size;
	size_t num_posterior_samples;
	bool sample_weights;
	ETerminationCriterion termination_criterion;
	size_t termination_num_steps;
	Real termination_remaining_contribution_threshold;
	Real ntimesh_factor;
	std::string sampling_strategy_str;
	Real initial_sampling_strategy_acceptance_threshold;
	size_t sampling_strategy_mcmc_numsteps;

	// Run-time variables
	RNG* rng;
	NestedSamplingStrategy* sampling_strategy;
	NestedSamplingStrategy* initial_sampling_strategy;
	NestedSamplingStrategy* sampling_from_strategy;
	Real information;
	SamplePopulation sampling_population;
	std::vector<Sample> posterior_reservoir;	//< The posterior samples arranged in a heap according to their weight
	Real current_acceptance_rate_ema;
	std::queue<Sample*> stored_samples;
	
	struct EvidenceSample
	{
		Real evidence;
		Real information;
		Real lXim1;
		Real lXi;
		Real lXip1;
	};
	std::vector<EvidenceSample> sampled_evidence;

	struct AsyncParams {
		RNG* rng;
	};
	std::vector<AsyncParams> async;
	struct AsyncTask {
		uint64 taskix;
		Sample* new_sample;
		Real min_llh;
		bool failure;
		Real acceptance_rate;
	};
	std::vector<AsyncTask> async_tasks;
	
	friend class NestedSamplingStrategy;
};

}
