#include "Utils.h"
#include "Clustering.h"
#include "Prior.h"
#include "RNG_SFMT.h"
#include "SamplerPT.h"
#include "SamplerPTChain.h"

namespace bcm {

static const size_t EMPIRICAL_STATS_SUBSAMPLING = 10;

SamplerPTChain::SamplerPTChain(SamplerPT* sampler)
	: sampler(sampler)
	, temperature(std::numeric_limits<Real>::quiet_NaN())
	, num_mcmc_moves_per_iteration(1)
	, rng(NULL)
	, lprior(-std::numeric_limits<Real>::infinity())
	, llh(-std::numeric_limits<Real>::infinity())
	, lpowerposterior(-std::numeric_limits<Real>::infinity())
	, attempted_mutate(0)
	, attempted_exchange(0)
	, accepted_mutate(0)
	, accepted_exchange(0)
	, empirical_n(0)
	, empirical_ix(0)
	, async_task(0)
	, async_burnin(false)
{
	current_var_values = VectorReal::Zero(sampler->num_variables);
	current_output_values = VectorReal::Zero(sampler->num_output_values);
	empirical_mean = VectorReal::Zero(sampler->num_variables);
	empirical_mean_nm1 = VectorReal::Zero(sampler->num_variables);
	empirical_variance = VectorReal::Zero(sampler->num_variables);
	empirical_variance_calc = VectorReal::Zero(sampler->num_variables);
	empirical_covariance = MatrixReal::Zero(sampler->num_variables, sampler->num_variables);
	empirical_correlation = MatrixReal::Zero(sampler->num_variables, sampler->num_variables);
	empirical_correlation_calc = MatrixReal::Zero(sampler->num_variables, sampler->num_variables);
}

SamplerPTChain::~SamplerPTChain()
{
	safe_delete(rng);
}

bool SamplerPTChain::Initialize(unsigned long rngseed)
{
	rng = new RNG_SFMT(rngseed);
	
	lprior = -std::numeric_limits<Real>::infinity();
	llh = -std::numeric_limits<Real>::infinity();
	lpowerposterior = -std::numeric_limits<Real>::infinity();
	current_var_values = VectorReal::Zero(sampler->num_variables);
	current_output_values = VectorReal::Zero(sampler->num_output_values);

	attempted_mutate = 0;
	attempted_exchange = 0;
	accepted_mutate = 0;
	accepted_exchange = 0;

	if (sampler->proposal_type == SamplerPT::PT_Scalar || sampler->proposal_type == SamplerPT::PT_AutoBlock) {
		// Start with scalar sampler
		sampler_blocks.resize(sampler->num_variables);
		for (size_t i = 0; i < sampler->num_variables; i++) {
			Real prior_var;
			if (!sampler->prior->EvaluateMarginalVariance(i, prior_var)) {
				// TODO - somehow come up with something reasonable?
				prior_var = 1.0;
			}
			sampler_blocks[i].variable_indices.resize(1, i);
			sampler_blocks[i].sigma = VectorReal::Constant(1, sqrt(prior_var));
			sampler_blocks[i].scale = 1.0;
			sampler_blocks[i].current_acceptance_rate_ema = sampler->target_acceptance_rate;
		}
	} else if (sampler->proposal_type == SamplerPT::PT_FullVector) {
		sampler_blocks.resize(1);
		sampler_blocks[0].variable_indices.resize(sampler->num_variables);
		sampler_blocks[0].sigma.resize(sampler->num_variables);
		for (size_t i = 0; i < sampler->num_variables; i++) {
			Real prior_var;
			if (!sampler->prior->EvaluateMarginalVariance(i, prior_var)) {
				// TODO - somehow come up with something reasonable?
				prior_var = 1.0;
			}
			sampler_blocks[0].variable_indices[i] = i;
			sampler_blocks[0].sigma[i] = sqrt(prior_var);
		}
		sampler_blocks[0].scale = 1.0 / (Real)sampler->num_variables;
		sampler_blocks[0].current_acceptance_rate_ema = sampler->target_acceptance_rate;
		sampler_blocks[0].mvn_sampler.SetCovariance(MatrixReal::Identity(sampler->num_variables, sampler->num_variables));
	} else {
		LOGERROR("Invalid proposal type");
		return false;
	}

	return true;
}

void SamplerPTChain::SetTemperature(Real temperature)
{
	this->temperature = temperature;
	lpowerposterior = lprior + temperature * llh;

	empirical_mean = VectorReal::Zero(sampler->num_variables);
	empirical_mean_nm1 = VectorReal::Zero(sampler->num_variables);
	empirical_variance = VectorReal::Zero(sampler->num_variables);
	empirical_variance_calc = VectorReal::Zero(sampler->num_variables);
	empirical_covariance = MatrixReal::Zero(sampler->num_variables, sampler->num_variables);
	empirical_correlation = MatrixReal::Zero(sampler->num_variables, sampler->num_variables);
	empirical_correlation_calc = MatrixReal::Zero(sampler->num_variables, sampler->num_variables);
	empirical_n = 0;
	empirical_ix = 0;
}

bool SamplerPTChain::AdaptProposal()
{
	// Calculate empirical variance
	for (size_t i = 0; i < sampler->num_variables; i++) {
		empirical_variance[i] = empirical_variance_calc[i] / (empirical_n - 1);
	}

	// Calculate empirical correlation
	for (size_t i = 0; i < sampler->num_variables; i++) {
		for (size_t j = 0; j < sampler->num_variables; j++) {
			if (j >= i) {
				empirical_covariance(i,j) = empirical_correlation_calc(i,j);
				empirical_correlation(i,j) = empirical_correlation_calc(i,j) / (sqrt(empirical_correlation_calc(i,i)) * sqrt(empirical_correlation_calc(j,j)));
			} else {
				empirical_covariance(i,j) = empirical_correlation_calc(j,i);
				empirical_correlation(i,j) = empirical_correlation_calc(j,i) / (sqrt(empirical_correlation_calc(i,i)) * sqrt(empirical_correlation_calc(j,j)));
			}
		}
	}
	empirical_covariance *= 1.0 / (Real)(empirical_n - 1);
	
	if (sampler->proposal_type == SamplerPT::PT_Scalar) {
		for (size_t i = 0; i < sampler->num_variables; i++) {
			// Make sure the proposal variance is not too small
			Real var = empirical_variance[i];
			Real prior_var;
			if (!sampler->prior->EvaluateMarginalVariance(i, prior_var)) {
				// TODO - somehow come up with something reasonable?
				prior_var = 1.0;
			}
			var = std::max(var, (Real)1e-6 * prior_var);
		
			sampler_blocks[i].variable_indices.resize(1, i);
			sampler_blocks[i].sigma = VectorReal::Constant(1, sqrt(var));
			sampler_blocks[i].scale = 1.0;
			sampler_blocks[i].current_acceptance_rate_ema = sampler->target_acceptance_rate;
		}
	} else if (sampler->proposal_type == SamplerPT::PT_FullVector) {
		for (size_t i = 0; i < sampler->num_variables; i++) {
			// Make sure the proposal variance is not too small
			Real var = empirical_variance[i];
			Real prior_var;
			if (!sampler->prior->EvaluateMarginalVariance(i, prior_var)) {
				// TODO - somehow come up with something reasonable?
				prior_var = 1.0;
			}
			var = std::max(var, (Real)1e-6 * prior_var);
			sampler_blocks[0].sigma[i] = sqrt(var);
		}
		sampler_blocks[0].scale = 1.0 / (Real)sampler->num_variables;
		sampler_blocks[0].current_acceptance_rate_ema = sampler->target_acceptance_rate;
	} else if (sampler->proposal_type == SamplerPT::PT_AutoBlock) {
		// Calculate distance for clustering
		MatrixReal distance = MatrixReal::Zero(empirical_correlation.cols(), empirical_correlation.cols());
		for (size_t i = 0; i < (size_t)empirical_correlation.rows(); i++) {
			for (size_t j = 0; j < (size_t)empirical_correlation.cols(); j++) {
				distance(i,j) = 1.0 - fabs(empirical_correlation(i,j));
			}
		}

		std::vector< std::set<size_t> > clusters;
		bcm::TreeCluster(distance, clusters, 0.5);
		sampler_blocks.clear();
		sampler_blocks.resize(clusters.size());
		for (size_t i = 0; i < clusters.size(); i++) {
			for (std::set<size_t>::iterator vi = clusters[i].begin(); vi != clusters[i].end(); ++vi) {
				sampler_blocks[i].variable_indices.push_back(*vi);
			}

			if (clusters[i].size() > 1) {
				MatrixReal cov = MatrixReal::Zero(clusters[i].size(), clusters[i].size());
				for (size_t x = 0; x < clusters[i].size(); x++) {
					for (size_t y = 0; y < clusters[i].size(); y++) {
						cov(x, y) = empirical_covariance(sampler_blocks[i].variable_indices[x], sampler_blocks[i].variable_indices[y]);
					}
				}
				sampler_blocks[i].mvn_sampler.SetCovariance(cov);
			} else {
				sampler_blocks[i].sigma = VectorReal::Zero(clusters[i].size());
				for (size_t vi = 0; vi < clusters[i].size(); vi++) {
					size_t ix = sampler_blocks[i].variable_indices[vi];

					// Make sure the proposal variance is not too small
					Real var = empirical_variance[ix];
					Real prior_var;
					if (!sampler->prior->EvaluateMarginalVariance(ix, prior_var)) {
						// TODO - somehow come up with something reasonable?
						prior_var = 1.0;
					}
					var = std::max(var, (Real)1e-6 * prior_var);
					sampler_blocks[i].sigma(vi) = sqrt(var);
				}
			}
			
			sampler_blocks[i].scale = 1.0;
			sampler_blocks[i].current_acceptance_rate_ema = sampler->target_acceptance_rate;
		}
	} else {
		LOGERROR("Invalid proposal type");
		return false;
	}

	// Reset empirical stat calculations
	empirical_mean = VectorReal::Zero(sampler->num_variables);
	empirical_mean_nm1 = VectorReal::Zero(sampler->num_variables);
	empirical_variance = VectorReal::Zero(sampler->num_variables);
	empirical_variance_calc = VectorReal::Zero(sampler->num_variables);
	empirical_correlation = MatrixReal::Zero(sampler->num_variables, sampler->num_variables);
	empirical_correlation_calc = MatrixReal::Zero(sampler->num_variables, sampler->num_variables);
	empirical_n = 0;
	empirical_ix = 0;

	return true;
}

void SamplerPTChain::SetMCMCMovesPerIteration(size_t num_moves)
{
	num_mcmc_moves_per_iteration = num_moves;
}

bool SamplerPTChain::FindStartingPosition()
{
	llh = -std::numeric_limits<Real>::infinity();
	lprior = -std::numeric_limits<Real>::infinity();
	lpowerposterior = -std::numeric_limits<Real>::infinity();

	static const size_t MAX_TRIES = 1000;
	for (size_t i = 0; i < MAX_TRIES; i++) {
		if (!sampler->prior->Sample(current_var_values.data(), rng)) {
			LOGERROR("Sampling from prior failed");
			return false;
		}

		if (!sampler->EvaluatePriorLikelihood(0, current_var_values.data(), lprior, llh)) {
			return false;
		}
		lpowerposterior = lprior + temperature * llh;

		if (lpowerposterior > -std::numeric_limits<Real>::infinity()) {
			if (!sampler->GetOutputValues(0, current_output_values.data())) {
				return false;
			}
			break;
		}
	}
	
	if (lpowerposterior == -std::numeric_limits<Real>::infinity()) {
		LOGERROR("Could not find starting position with power posterior != inf after %u tries", MAX_TRIES);
		return false;
	} else {
		return true;
	}
}

bool SamplerPTChain::MutateMove(bool burnin, size_t thread)
{
	// Sample each block separately
	for (size_t i = 0; i < sampler_blocks.size(); i++) {
		Block& b = sampler_blocks[i];

		// Update scale
		Real learn_rate = 1.0 + rng->GetReal() * sampler->proposal_scaling_learning_rate;
		if (b.current_acceptance_rate_ema < 0.952381 * sampler->target_acceptance_rate) {
			b.scale /= learn_rate;
			b.scale = std::max(b.scale, (Real)1e-6);
		} else if (b.current_acceptance_rate_ema > 1.05 * sampler->target_acceptance_rate) {
			b.scale *= learn_rate;
			b.scale = std::min(b.scale, (Real)100.0);
		}
		
		// Propose new parameters
		VectorReal new_values = current_var_values;
		if (b.variable_indices.size() == 1) {
			// Scalar
			for (size_t vi = 0; vi < b.variable_indices.size(); vi++) {
				size_t ix = b.variable_indices[vi];
				const Real lb = sampler->prior->GetLowerBound(ix);
				const Real ub = sampler->prior->GetUpperBound(ix);
				
				Real x = rng->GetNormal(current_var_values[ix], b.scale * b.sigma(vi));
				
				// Reflect for variables with bounded support, this is still a symmetrical proposal
				while (1) {
					if (x < lb) {
						x = lb + (lb - x);
					} else if (x > ub) {
						x = ub - (x - ub);
					} else {
						break;
					}
				}

				new_values[ix] = x;
			}
		} else {
			if (true) {
				VectorReal x = VectorReal::Zero(b.variable_indices.size());
				b.mvn_sampler.Get(rng, x);
				
				for (size_t vi = 0; vi < b.variable_indices.size(); vi++) {
					size_t ix = b.variable_indices[vi];
					const Real lb = sampler->prior->GetLowerBound(ix);
					const Real ub = sampler->prior->GetUpperBound(ix);

					x[vi] *= b.scale;
					x[vi] += current_var_values[ix];
				
					// Reflect for variables with bounded support, this is still a symmetrical proposal
					while (1) {
						if (x[vi] < lb) {
							x[vi] = lb + (lb - x[vi]);
						} else if (x[vi] > ub) {
							x[vi] = ub - (x[vi] - ub);
						} else {
							break;
						}
					}

					new_values[ix] = x[vi];
				}
			} else {
			}
		}

		// Evaluate prior, likelihood & posterior
		Real new_lprior = -std::numeric_limits<Real>::infinity();
		Real new_llh = -std::numeric_limits<Real>::infinity();
		Real new_lpowerposterior = -std::numeric_limits<Real>::infinity();
	
		if (!sampler->EvaluatePriorLikelihood(thread, new_values.data(), new_lprior, new_llh)) {
			return false;
		}
		new_lpowerposterior = new_lprior + temperature * new_llh;

		attempted_mutate++;

		// Do we accept the new move?
		bool accept = false;
		if (new_lpowerposterior > -std::numeric_limits<Real>::infinity()) {
			Real transition_probability = new_lpowerposterior - lpowerposterior;
			transition_probability = exp(transition_probability);
			transition_probability = std::min((Real)1.0, transition_probability);
	
			Real alpha = rng->GetReal();
			if (alpha < transition_probability) {
				accept = true;
			}
		}

		const Real ema_alpha = 2.0 / (sampler->proposal_scaling_ema_period + 1);
		if (accept) {
			accepted_mutate++;

			current_var_values = new_values;
			if (!sampler->GetOutputValues(thread, current_output_values.data())) {
				return false;
			}
			lprior = new_lprior;
			llh = new_llh;
			lpowerposterior = new_lpowerposterior;

			b.current_acceptance_rate_ema += (1.0 - b.current_acceptance_rate_ema) * ema_alpha;
		} else {
			b.current_acceptance_rate_ema += (0.0 - b.current_acceptance_rate_ema) * ema_alpha;
		}
		
		UpdateVariance();
	}

	return true;
}

void SamplerPTChain::MutateMoveAsync(bool burnin)
{
	async_burnin = burnin;
	TTask task = boost::bind(&SamplerPTChain::AsyncDoMutateMove, this, _1, _2);
	async_task = sampler->task_manager->AddTask(task, NULL);
}

bool SamplerPTChain::MutateMoveWait()
{
	sampler->task_manager->WaitTask(async_task);
	return async_result;
}

bool SamplerPTChain::ExchangeMove(bool burnin, SamplerPTChain& other)
{
	attempted_exchange++;

	SamplerPTChain& chain1 = *this;
	SamplerPTChain& chain2 = other;
	
	Real proposed_lpowerposteroir1 = chain1.temperature * chain2.llh + chain2.lprior;
	Real proposed_lpowerposteroir2 = chain2.temperature * chain1.llh + chain1.lprior;
	
	Real transition_probability = (proposed_lpowerposteroir1 + proposed_lpowerposteroir2) - (chain1.lpowerposterior + chain2.lpowerposterior);
	transition_probability = exp(transition_probability);
	transition_probability = std::min((Real)1.0, transition_probability);

	Real alpha = rng->GetReal();
	if (alpha < transition_probability) {
		accepted_exchange++;
		
		std::swap(chain1.current_var_values, chain2.current_var_values);
		std::swap(chain1.current_output_values, chain2.current_output_values);
		std::swap(chain1.llh, chain2.llh);
		std::swap(chain1.lprior, chain2.lprior);
		chain1.lpowerposterior = proposed_lpowerposteroir1;
		chain2.lpowerposterior = proposed_lpowerposteroir2;

		return true;
	} else {
		return false;
	}
}

void SamplerPTChain::LogStatistics()
{
	LOG("%11.7f | %12.5f | %14.5f",
		temperature,
		accepted_mutate / (Real)attempted_mutate,
		accepted_exchange / (Real)attempted_exchange);
}

void SamplerPTChain::LogProposalInfo()
{
	for (size_t i = 0; i < sampler_blocks.size(); i++) {
		Block& b = sampler_blocks[i];

		std::string ixs = std::to_string((uint64)b.variable_indices[0]);
		for (size_t vi = 1; vi < b.variable_indices.size(); vi++) {
			ixs += std::string(",") + std::to_string((uint64)b.variable_indices[vi]);
		}

		LOG("Block %u: %g - %s", i, b.scale, ixs.c_str());
	}
}

void SamplerPTChain::AsyncDoMutateMove(void* user, size_t thread)
{
	for (size_t i = 0; i < num_mcmc_moves_per_iteration; i++) {
		async_result = MutateMove(async_burnin, thread);
		if (!async_result) {
			break;
		}
	}
}

void SamplerPTChain::UpdateVariance()
{
	empirical_ix++;
	if (empirical_ix == EMPIRICAL_STATS_SUBSAMPLING) {
		empirical_ix = 0;
	} else {
		return;
	}

	empirical_n++;
	Real invN = 1.0 / (Real)empirical_n;
	
	empirical_mean_nm1 = empirical_mean;
	for (size_t i = 0; i < sampler->num_variables; i++) {
		const Real x = current_var_values[i];
		const Real d = x - empirical_mean[i];
		empirical_mean[i] += d * invN;
	}
	
	for (size_t i = 0; i < sampler->num_variables; i++) {
		const Real x = current_var_values[i];
		const Real f = (x - empirical_mean[i]) * (x - empirical_mean_nm1[i]);
		empirical_variance_calc[i] += f;
	}

	if (empirical_n > 1) {
		const Real ratio = (empirical_n-1) / (Real)empirical_n;
		for (size_t i = 0; i < sampler->num_variables; i++) {
			const Real dx = current_var_values[i] - empirical_mean_nm1[i];
			for (size_t j = i; j < sampler->num_variables; j++) {
				const Real dy = current_var_values[j] - empirical_mean_nm1[j];
				empirical_correlation_calc(i,j) += dx * dy * ratio;
			}
		}
	}
}

}
