#include "Utils.h"
#include "ProbabilityDistributions.h"
#include "RNG.h"
#include "UnivariateMarginalGMM.h"

namespace bcm {

UnivariateMarginalGMM::UnivariateMarginalGMM()
	: num_components(0)
{
}

UnivariateMarginalGMM::~UnivariateMarginalGMM()
{
}

bool UnivariateMarginalGMM::Initialize(boost::property_tree::ptree pt)
{
	try {
		num_components = pt.get<size_t>("<xmlattr>.num_components");
		if (num_components == 0) {
			LOGERROR("GMM must have at least 1 component");
			return false;
		}

		std::string proportions_str = pt.get<std::string>("<xmlattr>.proportions");
		std::string mu_str = pt.get<std::string>("<xmlattr>.mu");
		std::string sigma_str = pt.get<std::string>("<xmlattr>.sigma");

		std::vector<std::string> proportions_v;
		std::vector<std::string> mu_v;
		std::vector<std::string> sigma_v;

		tokenize(proportions_str, proportions_v, ",");
		tokenize(mu_str, mu_v, ",");
		tokenize(sigma_str, sigma_v, ",");

		if (proportions_v.size() != num_components) {
			LOGERROR("Inconsistent number of elements in proportions, should have %u elements", num_components);
			return false;
		}
		if (mu_v.size() != num_components) {
			LOGERROR("Inconsistent number of elements in mu, should have %u elements", num_components);
			return false;
		}
		if (sigma_v.size() != num_components) {
			LOGERROR("Inconsistent number of elements in sigma, should have %u elements", num_components);
			return false;
		}

		proportions.resize(num_components);
		mu.resize(num_components);
		sigma.resize(num_components);

		for (size_t i = 0; i < num_components; i++) {
			proportions(i) = atof(proportions_v[i].c_str());
			mu(i) = atof(mu_v[i].c_str());
			sigma(i) = atof(sigma_v[i].c_str());
		}
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing UnivariateMarginal: %s", e.what());
		return false;
	}

	return true;
}

bool UnivariateMarginalGMM::Sample(Real& value, RNG* rng) const
{
	// Sample a component
	size_t c = num_components - 1;
	Real p = rng->GetReal();
	Real t = 0;
	for (size_t i = 0; i < num_components - 1; i++) {
		t += proportions[i];
		if (p < t) {
			c = i;
			break;
		}
	}

	// Sample from that components
	value = rng->GetNormal(mu[c], sigma[c]);

	return true;
}

bool UnivariateMarginalGMM::EvaluatePDF(Real& p, const Real value) const
{
	p = 0;
	for (size_t i = 0; i < num_components; i++) {
		p += proportions[i] * PdfNormal(value, mu[i], sigma[i]);
	}
	return true;
}

bool UnivariateMarginalGMM::EvaluateLogPDF(Real& logp, const Real value) const
{
	Real p;
	if (!EvaluatePDF(p, value)) {
		return false;
	}
	logp = log(p);
	return true;
}

bool UnivariateMarginalGMM::EvaluateVariance(Real& var) const
{
	var = 0.0;
	for (size_t i = 0; i < num_components; i++) {
		var += proportions[i] * proportions[i] * sigma[i] * sigma[i];
	}
	return true;
}

bool UnivariateMarginalGMM::EvaluateVarianceLogTransformed(Real& var) const
{
	var = 0.0;
	for (size_t i = 0; i < num_components; i++) {
		Real thisvar = (exp(sigma[i] * sigma[i]) - 1.0) * exp(2.0 * mu[i] + sigma[i] * sigma[i]);
		var += proportions[i] * proportions[i] * thisvar;
	}
	return true;
}

}
