#include "Utils.h"
#include "LinearSolverEigenSparseLU.h"
#include "SBMLModel.h"

#include <cvode/cvode_impl.h>

LinearSolverEigenSparseLU::LinearSolverEigenSparseLU(SBMLModel* model)
	: LinearSolverSBML(model)
{
}

LinearSolverEigenSparseLU::~LinearSolverEigenSparseLU()
{
}

bool LinearSolverEigenSparseLU::Init(size_t N)
{
	if (N > std::numeric_limits<int>::max()) {
		LOGERROR("System too large");
		return false;
	}

	this->N = (int)N;
	M = Eigen::SparseMatrix<OdeReal>(this->N, this->N);
	SavedJ = Eigen::SparseMatrix<OdeReal>(this->N, this->N);

	return true;
}

#define PRINT_MATRICES 0

int LinearSolverEigenSparseLU::Setup(void* cvode_mem, int convfail, OdeReal* ypred, int* jcurPtr)
{
	CVodeMem cv_mem = (CVodeMem)cvode_mem;
	OdeReal gamma = cv_mem->cv_gamma;
	OdeReal dgamma = fabs((gamma / cv_mem->cv_gammap) - (OdeReal)(OdeReal)1.0);
	bool jbad = (cv_mem->cv_nst == 0) ||
				(cv_mem->cv_nst > nstlj + 50) ||
				((convfail == CV_FAIL_BAD_J) && (dgamma < 0.2)) ||
				(convfail == CV_FAIL_OTHER);

	// Evaluate or copy J
	if (jbad) {
		nstlj = cv_mem->cv_nst;
		M.setZero();
		EvaluateJacobian(ypred, Parameters);
		SavedJ = M;
		*jcurPtr = 1;
	} else {
		M = SavedJ;
		*jcurPtr = 0;
	}
	
#if PRINT_MATRICES
	Eigen::IOFormat CleanFmt(10, 0, ", ", "\n", "[", "]");
	std::fstream f;
	f.open("matrix.txt", std::fstream::out);

	f << "Jacobian (sparse)" << std::endl;
	f << M << std::endl;
	
	f << "Jacobian" << std::endl;
	f << MatrixReal(M).format(CleanFmt) << std::endl;
#endif
	
	// Construct M = I - gamma*J
	M *= -gamma;
	for (int i = 0; i < N; i++) {
		M.coeffRef(i,i) += (OdeReal)1.0;
	}
	M.makeCompressed();

#if PRINT_MATRICES
	f << "M (sparse)" << std::endl;
	f << M << std::endl;
	
	f << "M" << std::endl;
	f << MatrixReal(M).format(CleanFmt) << std::endl;
#endif

	solver.analyzePattern(M);
	solver.factorize(M);
	Eigen::ComputationInfo info = solver.info();
	if (info != Eigen::Success) {
		return 1;
	} else {
		MatrixReal perm = solver.rowsPermutation();

#if PRINT_MATRICES
		f << "row permutation" << std::endl;
		f << perm.format(CleanFmt) << std::endl;

		f.close();
#endif

		return 0;
	}
}

int LinearSolverEigenSparseLU::Solve(void* cvode_mem, OdeReal* b)
{
	CVodeMem cv_mem = (CVodeMem)cvode_mem;

#if 1
	Eigen::Map< Eigen::Matrix< OdeReal, Eigen::Dynamic, 1 >  > bv(b, N);
	bv = solver.solve(bv);
	
	OdeReal gamrat = cv_mem->cv_gamrat;
	if ((cv_mem->cv_lmm == CV_BDF) && (gamrat != (OdeReal)1.0)) {
		bv *= (OdeReal)2.0 / ((OdeReal)(OdeReal)1.0 + gamrat);
	}
#else
	VectorReal x(N), bv(N);
	for (size_t i = 0; i < N; i++) {
		bv(i) = b[i];
	}
	x = solver.solve(bv);
	for (size_t i = 0; i < N; i++) {
		b[i] = x(i);
	}
#endif
	return 0;
}

void LinearSolverEigenSparseLU::EvaluateJacobian(const OdeReal* species, const OdeReal* parameters)
{
	M.coeffRef(0,0) = -(parameters[0]*(OdeReal)1.0); // j=0
	M.coeffRef(1,0) = +((+(parameters[21]*(OdeReal)1.0))/(1+exp((parameters[30]*(log(species[4])-parameters[32]))))); // j=0

	M.coeffRef(1,1) = -(parameters[10]*(OdeReal)1.0); // j=1
	M.coeffRef(14,1) = +(parameters[35]*(OdeReal)1.0); // j=1

	M.coeffRef(2,2) = -(parameters[10]*(OdeReal)1.0); // j=2
	M.coeffRef(15,2) = +(parameters[35]*(OdeReal)1.0); // j=2

	M.coeffRef(3,3) = -(parameters[10]*(OdeReal)1.0); // j=3
	M.coeffRef(16,3) = +(parameters[35]*(OdeReal)1.0); // j=3

	M.coeffRef(1,4) = +(-(+exp((parameters[30]*(log(species[4])-parameters[32])))*((parameters[30]*(OdeReal)1.0/(species[4]))))*(parameters[15]+(parameters[21]*species[0]))/((1+exp((parameters[30]*(log(species[4])-parameters[32]))))*(1+exp((parameters[30]*(log(species[4])-parameters[32])))))); // j=4
	M.coeffRef(4,4) = -(parameters[5]*(OdeReal)1.0); // j=4
	M.coeffRef(17,4) = +(-(+exp((parameters[29]*(log(species[4])-parameters[31])))*((parameters[29]*(OdeReal)1.0/(species[4]))))*(parameters[11]+(parameters[20]*species[15]))/((1+exp((parameters[29]*(log(species[4])-parameters[31]))))*(1+exp((parameters[29]*(log(species[4])-parameters[31])))))); // j=4

	M.coeffRef(4,5) = +(parameters[35]*(OdeReal)1.0); // j=5
	M.coeffRef(5,5) = -(parameters[10]*(OdeReal)1.0); // j=5

	M.coeffRef(3,6) = +(+(parameters[26]*(OdeReal)1.0)); // j=6
	M.coeffRef(6,6) = -(parameters[6]*(OdeReal)1.0); // j=6

	M.coeffRef(6,7) = +(parameters[35]*(OdeReal)1.0); // j=7
	M.coeffRef(7,7) = -(parameters[10]*(OdeReal)1.0); // j=7

	M.coeffRef(8,8) = -(parameters[1]*(OdeReal)1.0); // j=8
	M.coeffRef(16,8) = -(+(-(+exp((parameters[33]*(log(species[8])-parameters[34])))*((parameters[33]*(OdeReal)1.0/(species[8]))))*(parameters[9]*species[16])/((1+exp((parameters[33]*(log(species[8])-parameters[34]))))*(1+exp((parameters[33]*(log(species[8])-parameters[34]))))))); // j=8

	M.coeffRef(8,9) = +(parameters[35]*(OdeReal)1.0); // j=9
	M.coeffRef(9,9) = -(parameters[10]*(OdeReal)1.0); // j=9

	M.coeffRef(10,10) = -(parameters[2]*(OdeReal)1.0); // j=10

	M.coeffRef(11,11) = -(parameters[3]*(OdeReal)1.0); // j=11

	M.coeffRef(10,12) = +(parameters[35]*(OdeReal)1.0); // j=12
	M.coeffRef(12,12) = -(parameters[10]*(OdeReal)1.0); // j=12

	M.coeffRef(11,13) = +(parameters[35]*(OdeReal)1.0); // j=13
	M.coeffRef(13,13) = -(parameters[10]*(OdeReal)1.0); // j=13

	M.coeffRef(5,14) = +(+(parameters[22]*(OdeReal)1.0)); // j=14
	M.coeffRef(7,14) = +(+(parameters[23]*(OdeReal)1.0)); // j=14
	M.coeffRef(9,14) = +(+(parameters[24]*(OdeReal)1.0)); // j=14
	M.coeffRef(12,14) = +(+(parameters[25]*(OdeReal)1.0)); // j=14
	M.coeffRef(14,14) = -(parameters[4]*(OdeReal)1.0); // j=14

	M.coeffRef(15,15) = -(parameters[8]*(OdeReal)1.0); // j=15
	M.coeffRef(17,15) = +((+(parameters[20]*(OdeReal)1.0))/(1+exp((parameters[29]*(log(species[4])-parameters[31]))))); // j=15

	M.coeffRef(2,16) = +(+(parameters[27]*(OdeReal)1.0)); // j=16
	M.coeffRef(13,16) = +(+(parameters[28]*(OdeReal)1.0)); // j=16
	M.coeffRef(16,16) = -(+(parameters[7]*(OdeReal)1.0)+((parameters[9]*(OdeReal)1.0)/(1+exp((parameters[33]*(log(species[8])-parameters[34])))))); // j=16

	M.coeffRef(0,17) = +(parameters[35]*(OdeReal)1.0); // j=17
	M.coeffRef(17,17) = -(parameters[10]*(OdeReal)1.0); // j=17
}
