#include "Utils.h"
#include "LinearSolverMKL.h"
#include "SBMLModel.h"

#if USE_MKL

#include <cvode/cvode_impl.h>
#include <mkl.h>

//#pragma(comment, "mkl_intel_lp64.lib")
//#pragma(comment, "mkl_core.lib")
//#pragma(comment, "mkl_blas95_lp64.lib")
//#pragma(comment, "mkl_lapack95_lp64.lib")
//#pragma(comment, "mkl_sequential.lib")

LinearSolverMKL::LinearSolverMKL(SBMLModel* model)
	: LinearSolverSBML(model)
{
}

LinearSolverMKL::~LinearSolverMKL()
{
}

bool LinearSolverMKL::Init(size_t N)
{
	this->N = N;
	M = OdeMatrixReal(N, N);
	SavedJ = OdeMatrixReal(N, N);
	pivots.resize(N, 0);

	return true;
}

int LinearSolverMKL::Setup(void* cvode_mem, int convfail, OdeReal* ypred, int* jcurPtr)
{
	CVodeMem cv_mem = (CVodeMem)cvode_mem;
	OdeReal gamma = cv_mem->cv_gamma;
	OdeReal dgamma = fabs((gamma / cv_mem->cv_gammap) - (OdeReal)1.0);
	bool jbad = (cv_mem->cv_nst == 0) ||
				(cv_mem->cv_nst > nstlj + 50) ||
				((convfail == CV_FAIL_BAD_J) && (dgamma < 0.2)) ||
				(convfail == CV_FAIL_OTHER);

	// Evaluate or copy J
	if (jbad) {
		nstlj = cv_mem->cv_nst;
		M.setZero();
		EvaluateJacobian(ypred, Parameters);
		SavedJ = M;
		*jcurPtr = 1;
	} else {
		M = SavedJ;
		*jcurPtr = 0;
	}
	
	// Construct M = I - gamma*J
	M *= -gamma;
	for (size_t i = 0; i < N; i++) {
		M.coeffRef(i,i) += (OdeReal)1.0;
	}
	
	solver.compute(M);
	//LAPACKE_dgetrf(LAPACK_COL_MAJOR, N, N, M.data(), N, pivots.data());

	return 0;
}

int LinearSolverMKL::Solve(void* cvode_mem, OdeReal* b)
{
	CVodeMem cv_mem = (CVodeMem)cvode_mem;
	
	Eigen::Map< Eigen::Matrix< OdeReal, Eigen::Dynamic, 1 >  > bv(b, N);
	bv = solver.solve(bv);

	//LAPACKE_dgetrs(LAPACK_COL_MAJOR, 'N', N, 1, M.data(), N, pivots.data(), b, N);
	
	OdeReal gamrat = cv_mem->cv_gamrat;
	if ((cv_mem->cv_lmm == CV_BDF) && (gamrat != (OdeReal)1.0)) {
		OdeReal f = 2.0 / ((OdeReal)1.0 + gamrat);
		for (size_t i = 0; i < N; i++) {
			b[i] *= f;
		}
	}

	return 0;
}

void LinearSolverMKL::EvaluateJacobian(const OdeReal* species, const OdeReal* parameters)
{
	M.coeffRef(0,0) = -(parameters[0]*(OdeReal)1.0); // j=0
	M.coeffRef(1,0) = +((+(parameters[29]*(OdeReal)1.0))/(1+exp((parameters[38]*(log(species[4])-parameters[40]))))); // j=0

	M.coeffRef(1,1) = -(parameters[14]*(OdeReal)1.0); // j=1
	M.coeffRef(14,1) = +(parameters[47]*(OdeReal)1.0); // j=1

	M.coeffRef(2,2) = -(parameters[18]*(OdeReal)1.0); // j=2
	M.coeffRef(15,2) = +(parameters[51]*(OdeReal)1.0); // j=2

	M.coeffRef(3,3) = -(parameters[17]*(OdeReal)1.0); // j=3
	M.coeffRef(16,3) = +(parameters[50]*(OdeReal)1.0); // j=3

	M.coeffRef(1,4) = +(-(+exp((parameters[38]*(log(species[4])-parameters[40])))*((parameters[38]*(OdeReal)1.0/(species[4]))))*(parameters[23]+(parameters[29]*species[0]))/((1+exp((parameters[38]*(log(species[4])-parameters[40]))))*(1+exp((parameters[38]*(log(species[4])-parameters[40])))))); // j=4
	M.coeffRef(4,4) = -(parameters[5]*(OdeReal)1.0); // j=4
	M.coeffRef(17,4) = +(-(+exp((parameters[37]*(log(species[4])-parameters[39])))*((parameters[37]*(OdeReal)1.0/(species[4]))))*(parameters[19]+(parameters[28]*species[15]))/((1+exp((parameters[37]*(log(species[4])-parameters[39]))))*(1+exp((parameters[37]*(log(species[4])-parameters[39])))))); // j=4

	M.coeffRef(4,5) = +(parameters[48]*(OdeReal)1.0); // j=5
	M.coeffRef(5,5) = -(parameters[15]*(OdeReal)1.0); // j=5

	M.coeffRef(3,6) = +(+(parameters[34]*(OdeReal)1.0)); // j=6
	M.coeffRef(6,6) = -(parameters[6]*(OdeReal)1.0); // j=6

	M.coeffRef(6,7) = +(parameters[49]*(OdeReal)1.0); // j=7
	M.coeffRef(7,7) = -(parameters[16]*(OdeReal)1.0); // j=7

	M.coeffRef(8,8) = -(parameters[1]*(OdeReal)1.0); // j=8
	M.coeffRef(16,8) = -(+(-(+exp((parameters[41]*(log(species[8])-parameters[42])))*((parameters[41]*(OdeReal)1.0/(species[8]))))*(parameters[9]*species[16])/((1+exp((parameters[41]*(log(species[8])-parameters[42]))))*(1+exp((parameters[41]*(log(species[8])-parameters[42]))))))); // j=8

	M.coeffRef(8,9) = +(parameters[44]*(OdeReal)1.0); // j=9
	M.coeffRef(9,9) = -(parameters[11]*(OdeReal)1.0); // j=9

	M.coeffRef(10,10) = -(parameters[2]*(OdeReal)1.0); // j=10

	M.coeffRef(11,11) = -(parameters[3]*(OdeReal)1.0); // j=11

	M.coeffRef(10,12) = +(parameters[45]*(OdeReal)1.0); // j=12
	M.coeffRef(12,12) = -(parameters[12]*(OdeReal)1.0); // j=12

	M.coeffRef(11,13) = +(parameters[46]*(OdeReal)1.0); // j=13
	M.coeffRef(13,13) = -(parameters[13]*(OdeReal)1.0); // j=13

	M.coeffRef(5,14) = +(+(parameters[30]*(OdeReal)1.0)); // j=14
	M.coeffRef(7,14) = +(+(parameters[31]*(OdeReal)1.0)); // j=14
	M.coeffRef(9,14) = +(+(parameters[32]*(OdeReal)1.0)); // j=14
	M.coeffRef(12,14) = +(+(parameters[33]*(OdeReal)1.0)); // j=14
	M.coeffRef(14,14) = -(parameters[4]*(OdeReal)1.0); // j=14

	M.coeffRef(15,15) = -(parameters[8]*(OdeReal)1.0); // j=15
	M.coeffRef(17,15) = +((+(parameters[28]*(OdeReal)1.0))/(1+exp((parameters[37]*(log(species[4])-parameters[39]))))); // j=15

	M.coeffRef(2,16) = +(+(parameters[35]*(OdeReal)1.0)); // j=16
	M.coeffRef(13,16) = +(+(parameters[36]*(OdeReal)1.0)); // j=16
	M.coeffRef(16,16) = -(+(parameters[7]*(OdeReal)1.0)+((parameters[9]*(OdeReal)1.0)/(1+exp((parameters[41]*(log(species[8])-parameters[42])))))); // j=16

	M.coeffRef(0,17) = +(parameters[43]*(OdeReal)1.0); // j=17
	M.coeffRef(17,17) = -(parameters[10]*(OdeReal)1.0); // j=17
}

#endif
