#pragma once

class CVODESolver;
class ExperimentalCondition;
class SBMLAssignmentRule;
class SBMLModelParameters;
class SBMLReaction;
class SBMLSpecies;

#include <sbml/SBMLTypes.h>
#include "VariableSet.h"
#include "LinearSolverSBML.h"

class SBMLModel
{
public:
	SBMLModel(size_t numthreads);
	~SBMLModel();

	bool LoadSBML(const std::string& fn);
	bool SetVariableSet(const bcm::VariableSet* variables);
	bool Initialize();
	void DumpODEStatistics(const std::string& base_filename);

	std::string GenerateCode();
	std::string GenerateJacobianCode(size_t i, size_t j);
	
	void ResetForcedVariables(size_t threadix);
	bool SetForcedVariable(const std::string& variable_name, Real value, size_t threadix);
	bool Simulate(const Real* variables, const std::vector<Real>& timepoints, size_t threadix);
	bool GetOutput(size_t species_ix, VectorReal& output, size_t threadix);
	bool GetOutput(const std::string& species_full_name, VectorReal& output, size_t threadix);

	void GetParameters(std::vector<std::string>& parameters) const;
	size_t GetNumSimulatedSpecies() const;
	const std::string& GetSimulatedSpeciesName(size_t i) const;
	std::string GetSimulatedSpeciesFullName(size_t i) const;
	void GetSimulatedSpeciesFullNames(std::vector<std::string>& species_names) const;
	const SBMLSpecies* GetSpecies(const std::string& id) const;
	const SBMLSpecies* GetSpeciesFromFullName(const std::string& id) const;
	SBMLModelParameters* GetParameterSet();

private:
	bool CalculateDerivative(OdeReal t, const OdeReal* y, OdeReal* dydt, void* user) const;
	bool CalculateJacobian(OdeReal t, const OdeReal* y, const OdeReal* dydt, OdeReal** jac, void* user) const;
	void CalculateAssignments(const Real* y, const Real* variables, Real* out) const;

	SBMLDocument* document;
	const Model* model;
	const bcm::VariableSet* variables;
	
	std::map<std::string, SBMLSpecies*> Species;
	std::map<std::string, SBMLReaction*> Reactions;
	std::vector<std::string> SimulatedSpecies;
	std::vector<SBMLAssignmentRule*> AssignmentRules;
	std::vector<size_t> AssignmentRulesTargetIx;
	std::vector<bool> SpeciesIsAssigned;

	struct ParallelSolver {
		ParallelSolver() : solver(NULL), linear_solver(NULL), parameters(NULL), variables(NULL) {}
		CVODESolver* solver;
		LinearSolverSBML* linear_solver;
		SBMLModelParameters* parameters;
		std::vector<Real> y;
		const Real* variables;
		VectorReal transformed_variables;
		OdeVectorReal transformed_variables_use_in_ode;
		MatrixReal trajectories;
	};
	std::vector<ParallelSolver> Solvers;
};
