#include "Utils.h"
#include "LikelihoodDataFile.h"
#include "LikelihoodExperiment.h"

LikelihoodDataFile::LikelihoodDataFile(size_t numthreads)
	: num_threads(numthreads)
{
}

LikelihoodDataFile::~LikelihoodDataFile()
{
	for (std::vector<LikelihoodExperiment*>::iterator ei = Experiments.begin(); ei != Experiments.end(); ++ei) {
		delete *ei;
	}
}

bool LikelihoodDataFile::Load(const boost::property_tree::ptree& xml_node)
{
	bcm::NetCDFDataFile file;

	try {
		Filename = xml_node.get<std::string>("<xmlattr>.file");

		if (!file.Open(Filename, netCDF::NcFile::read)) {
			return false;
		}

		BOOST_FOREACH(const boost::property_tree::ptree::value_type& var, xml_node.get_child("")) {
			if (var.first == "experiment") {
				LikelihoodExperiment* experiment = new LikelihoodExperiment(num_threads);
				if (!experiment->Load(var.second, &file)) {
					return false;
				}
				Experiments.push_back(experiment);
			}
		}

		file.Close();
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing likelihood file: %s", e.what());
		return false;
	}

	return true;
}

bool LikelihoodDataFile::SetVariableSet(const bcm::VariableSet* variables)
{
	for (std::vector<LikelihoodExperiment*>::iterator ei = Experiments.begin(); ei != Experiments.end(); ++ei) {
		if (!(*ei)->SetVariableSet(variables)) {
			return false;
		}
	}

	return true;
}

bool LikelihoodDataFile::Initialize()
{
	for (std::vector<LikelihoodExperiment*>::iterator ei = Experiments.begin(); ei != Experiments.end(); ++ei) {
		if (!(*ei)->Initialize()) {
			return false;
		}
	}

	return true;
}

void LikelihoodDataFile::SetNumModelSpecies(size_t n)
{
	for (std::vector<LikelihoodExperiment*>::iterator ei = Experiments.begin(); ei != Experiments.end(); ++ei) {
		(*ei)->SetNumModelSpecies(n);
	}
}

bool LikelihoodDataFile::SetupAdditionalOutput(bcm::NetCDFDataFile* file, SBMLModel* model, size_t sample_count)
{
	for (std::vector<LikelihoodExperiment*>::iterator ei = Experiments.begin(); ei != Experiments.end(); ++ei) {
		if (!(*ei)->SetupAdditionalOutput(file, model, sample_count)) {
			return false;
		}
	}
	return true;
}

void LikelihoodDataFile::HandleAdditionalOutput(bcm::NetCDFDataFile* file, SBMLModel* model, size_t sample_ix, const Real* values, size_t& value_ix)
{
	size_t ix = 0;
	for (std::vector<LikelihoodExperiment*>::iterator ei = Experiments.begin(); ei != Experiments.end(); ++ei) {
		size_t experiment_numvalues = 0;
		(*ei)->HandleAdditionalOutput(file, model, sample_ix, values + ix, experiment_numvalues);
		ix += experiment_numvalues;
	}
}

bool LikelihoodDataFile::EvaluateLogProbability(size_t threadix, const Real* values, SBMLModel* model, Real& logp)
{
	logp = 0.0;
	for (std::vector<LikelihoodExperiment*>::iterator ei = Experiments.begin(); ei != Experiments.end(); ++ei) {
		Real experiment_logp = std::numeric_limits<Real>::quiet_NaN();
		if (!(*ei)->EvaluateLogProbability(threadix, values, model,  experiment_logp)) {
			return false;
		}
		logp += experiment_logp;

		if (logp == -std::numeric_limits<Real>::infinity()) {
			// No need to continue
			break;
		}
	}

	return true;
}

bool LikelihoodDataFile::GetOutput(size_t threadix, Real* output_values, size_t& numvalues)
{
	if (output_values) {
		size_t ix = 0;
		for (std::vector<LikelihoodExperiment*>::iterator ei = Experiments.begin(); ei != Experiments.end(); ++ei) {
			size_t experiment_numvalues = 0;
			if (!(*ei)->GetOutput(threadix, output_values + ix, experiment_numvalues)) {
				return false;
			}
			ix += experiment_numvalues;
		}
	} else {
		for (std::vector<LikelihoodExperiment*>::iterator ei = Experiments.begin(); ei != Experiments.end(); ++ei) {
			size_t experiment_numvalues = 0;
			if (!(*ei)->GetOutput(threadix, output_values, experiment_numvalues)) {
				return false;
			}
			numvalues += experiment_numvalues;
		}
	}

	return true;
}
