#include "Utils.h"
#include "LikelihoodPart.h"
#include "ProbabilityDistributions.h"
#include "VariableSet.h"

#include <boost/lexical_cast.hpp>

LikelihoodPart::LikelihoodPart()
{
}

LikelihoodPart::~LikelihoodPart()
{
}

bool LikelihoodPart::Load(const boost::property_tree::ptree& xml_node, const bcm::NetCDFDataFile* datafile)
{
	if (!datafile) {
		LOGERROR("Must specify a data file");
		return false;
	}

	// Parse XML
	try {
		SpeciesName = xml_node.get<std::string>("<xmlattr>.species_name");
		std::string data_name  = xml_node.get<std::string>("<xmlattr>.data_name");
		size_t name_ix = data_name.find_last_of('/');
		if (name_ix == std::string::npos) {
			DataGroup = "";
			DataName = data_name;
		} else {
			DataGroup = data_name.substr(0, name_ix);
			DataName = data_name.substr(name_ix+1);
		}
		
		std::string likelihood_function = xml_node.get<std::string>("<xmlattr>.likelihood_function");

		if (likelihood_function == "normal") {
			LikelihoodFunction = LF_Normal;
			VarianceParameterName = xml_node.get<std::string>("<xmlattr>.normal_sigma");
		} else if (likelihood_function == "studentst") {
			LikelihoodFunction = LF_StudentsT;
			VarianceParameterName = xml_node.get<std::string>("<xmlattr>.t_sigma");
			T_Nu = xml_node.get<Real>("<xmlattr>.t_nu");
			ScaleParameterName = xml_node.get<std::string>("<xmlattr>.t_scale", "");
		} else if (likelihood_function == "delta") {
			LikelihoodFunction = LF_Delta;
		} else {
			LOGERROR("Unknown likelihood function %s in likelihood for trajectory of species %s", likelihood_function.c_str(), SpeciesName.c_str());
			return false;
		}
	} catch (boost::property_tree::ptree_error &e) {
		LOGERROR("Error parsing likelihood file: %s", e.what());
		return false;
	}

	if (!ParseDataName(datafile)) {
		return false;
	}

	return true;
}

bool LikelihoodPart::SetVariableSet(const bcm::VariableSet* variables)
{
	switch (LikelihoodFunction) {
	case LF_Normal:
		VarianceParameterIx = variables->GetVariableIndex(VarianceParameterName);
		if (VarianceParameterIx == std::numeric_limits<size_t>::max()) {
			LOGERROR("Variance paramater \"%s\" not in variable set", VarianceParameterName.c_str());
			return false;
		}
		VarianceParameterLog = variables->IsVariableLogSpace(VarianceParameterIx);
		break;
		
	case LF_StudentsT:
		VarianceParameterIx = variables->GetVariableIndex(VarianceParameterName);
		if (VarianceParameterIx == std::numeric_limits<size_t>::max()) {
			LOGERROR("Variance paramater \"%s\" not in variable set", VarianceParameterName.c_str());
			return false;
		}
		VarianceParameterLog = variables->IsVariableLogSpace(VarianceParameterIx);
		if (!ScaleParameterName.empty()) {
			ScaleParameterIx = variables->GetVariableIndex(ScaleParameterName);
			if (ScaleParameterIx == std::numeric_limits<size_t>::max()) {
				LOGERROR("Scale paramater \"%s\" not in variable set", ScaleParameterName.c_str());
				return false;
			}
		} else {
			ScaleParameterIx = std::numeric_limits<size_t>::max();
		}
		break;

	default:
		break;
	}

	return true;
}

bool LikelihoodPart::GetRequiredTimepoints(std::vector<Real>& timepoints)
{
	timepoints.clear();
	return true;
}

bool LikelihoodPart::ParseDataName(const bcm::NetCDFDataFile* datafile)
{
	size_t indexing_start = DataName.find_first_of('[');
	if (indexing_start == DataName.npos) {
		// No indexing specified
		DataNameWithoutIndex = DataName;
		return true;
	}

	size_t indexing_end = DataName.find_first_of(']');
	if (indexing_end == DataName.npos) {
		LOGERROR("Syntax error in DataName: missing closing bracket for indexing");
		return false;
	}

	DataNameWithoutIndex = DataName.substr(0, indexing_start);

	std::string indexingstr = DataName.substr(indexing_start+1, indexing_end - indexing_start - 1);
	std::vector<std::string> indexing_parts;
	bcm::tokenize(indexingstr, indexing_parts, ",");

	for (size_t i = 0; i < indexing_parts.size(); i++) {
		if (indexing_parts[i] == ":") {
		} else {
			if (indexing_parts[i].front() == '\'' && indexing_parts[i].back() == '\'') {
				std::string val = indexing_parts[i].substr(1, indexing_parts[i].size() - 2);
				std::string dimname;
				if (!datafile->GetDimensionName(DataGroup, DataNameWithoutIndex, i, dimname)) {
					return false;
				}

				size_t ix;
				if (!datafile->GetDimensionIx(DataGroup, dimname, val, &ix)) {
					return false;
				}
				DataIndices.push_back(ix);
			} else {
				size_t ix;
				try {
					ix = boost::lexical_cast<size_t>(indexing_parts[i]);
				} catch( boost::bad_lexical_cast const& ) {
					LOGERROR("Error converting index to a number");
					return false;
				}
				DataIndices.push_back(ix);
			}
		}
	}

	return true;
}

bool LikelihoodPart::LookupDataValue1D(const bcm::NetCDFDataFile* datafile, size_t ix, Real* value)
{
	size_t dimcount;
	if (!datafile->GetDimensionCount(DataGroup, DataNameWithoutIndex, &dimcount)) {
		return false;
	}

	if (dimcount != DataIndices.size() + 1) {
		LOGERROR("Incorrect number of indices for variable %s/%s", DataGroup.c_str(), DataName.c_str());
		return false;
	}

	if (DataIndices.empty()) {
		return datafile->GetValue(DataGroup, DataNameWithoutIndex, ix, value);
	} else if (DataIndices.size() == 1) {
		return datafile->GetValue(DataGroup, DataNameWithoutIndex, DataIndices[0], ix, value);
	} else if (DataIndices.size() == 2) {
		return datafile->GetValue(DataGroup, DataNameWithoutIndex, DataIndices[0], DataIndices[1], ix, value);
	} else {
		return false;
	}
}

bool LikelihoodPart::LookupDataValue2D(const bcm::NetCDFDataFile* datafile, size_t ix1, size_t ix2, Real* value)
{
	size_t dimcount;
	if (!datafile->GetDimensionCount(DataGroup, DataNameWithoutIndex, &dimcount)) {
		return false;
	}

	if (dimcount != DataIndices.size() + 2) {
		LOGERROR("Incorrect number of indices for variable %s/%s", DataGroup.c_str(), DataName.c_str());
		return false;
	}

	if (DataIndices.empty()) {
		return datafile->GetValue(DataGroup, DataNameWithoutIndex, ix1, ix2, value);
	} else if (DataIndices.size() == 1) {
		return datafile->GetValue(DataGroup, DataNameWithoutIndex, DataIndices[0], ix1, ix2, value);
	} else {
		return false;
	}
}

void LikelihoodPart::PreEvaluateLogPdf(const Real* values, Real& optim)
{
	if (LikelihoodFunction == LF_StudentsT) {
		Real var = values[VarianceParameterIx];
		if (VarianceParameterLog) {
			var = pow(10.0, var);
		}
		optim = bcm::LogPdfT_CalcC(var, T_Nu);
	}
}

Real LikelihoodPart::EvaluateLogPdf(const Real* values, Real modeled, Real observed, Real* optim)
{
	Real var = values[VarianceParameterIx];
	if (VarianceParameterLog) {
		var = pow(10.0, var);
	}

	switch (LikelihoodFunction) {
	case LF_Normal:
		return bcm::LogPdfNormal(observed, modeled, var);

	case LF_StudentsT:
		if (optim == NULL) {
			if (ScaleParameterIx == std::numeric_limits<size_t>::max()) {
				return bcm::LogPdfT(observed, modeled, var, T_Nu);
			} else {
				return bcm::LogPdfT(observed, modeled * values[ScaleParameterIx], var, T_Nu);
			}
		} else {
			if (ScaleParameterIx == std::numeric_limits<size_t>::max()) {
				return bcm::LogPdfT(observed, modeled, var, T_Nu, *optim);
			} else {
				return bcm::LogPdfT(observed, modeled * values[ScaleParameterIx], var, T_Nu, *optim);
			}
		}

	default:
		LOGERROR("");
		return 0.0;
	}
}
