#include "Utils.h"
#include "NetCDFDataFile.h"

#ifdef BCM_INCLUDE_NETCDF

namespace bcm {

NetCDFDataFile::NetCDFDataFile()
	: File(NULL)
{
}

NetCDFDataFile::~NetCDFDataFile()
{
	bcm::safe_delete(File);
}

bool NetCDFDataFile::Open(const std::string& filename, netCDF::NcFile::FileMode mode)
{
	try {
		bcm::safe_delete(File);
		File = new netCDF::NcFile(filename, mode);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error reading netcdf file \"%s\": %s", filename.c_str(), e.what());
		if (e.what() == "Unknown error") {
			LOGERROR("(perhaps file does not exist?)");
		}
		return false;
	}
}

void NetCDFDataFile::Close()
{
	try {
		bcm::safe_delete(File);
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error closing netcdf file: %s", e.what());
	}
}

bool NetCDFDataFile::IsOpen()
{
	return File != NULL;
}

bool NetCDFDataFile::GetValue(const std::string& group_name, const std::string& variable_name, size_t dim1ix, Real* value) const
{
	ASSERT(value);

	*value = -std::numeric_limits<Real>::quiet_NaN();

	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);

		std::vector<size_t> ix(1);
		ix[0] = dim1ix;
		var.getVar(ix, value);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%ul]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValue(const std::string& group_name, const std::string& variable_name, size_t dim1ix, std::string& value) const
{
	value.clear();

	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);

		std::vector<size_t> ix(1);
		ix[0] = dim1ix;
		char* str;
		var.getVar(ix, &str);
		value = str;
		free(str);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%ul]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValue(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, Real* value) const
{
	ASSERT(value);

	*value = -std::numeric_limits<Real>::quiet_NaN();

	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);

		std::vector<size_t> ix(2);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		var.getVar(ix, value);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%ul,%ul]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, dim2ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValue(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, size_t dim3ix, Real* value) const
{
	ASSERT(value);

	*value = -std::numeric_limits<Real>::quiet_NaN();

	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);

		std::vector<size_t> ix(3);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		ix[2] = dim3ix;
		var.getVar(ix, value);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%ul,%ul,%ul]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, dim2ix, dim3ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValue(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, size_t dim3ix, size_t dim4ix, Real* value) const
{
	ASSERT(value);

	*value = -std::numeric_limits<Real>::quiet_NaN();

	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);

		std::vector<size_t> ix(4);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		ix[2] = dim3ix;
		ix[3] = dim4ix;
		var.getVar(ix, value);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%ul,%ul,%ul,%ul]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, dim2ix, dim3ix, dim4ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValue(const std::string& group_name, const std::string& variable_name, const std::string& dim1value, Real* value) const
{
	ASSERT(value);

	*value = -std::numeric_limits<Real>::quiet_NaN();

	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);

		std::vector<size_t> ix(1);
		if (!GetDimensionIx(group, var.getDim(0), dim1value, &ix[0])) {
			return false;
		} else {
			var.getVar(ix, value);
			return true;
		}
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%s]: %s", group_name.c_str(), variable_name.c_str(), dim1value.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValue(const std::string& group_name, const std::string& variable_name, const std::string& dim1value, const std::string& dim2value, Real* value) const
{
	ASSERT(value);

	*value = -std::numeric_limits<Real>::quiet_NaN();

	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);

		std::vector<size_t> ix(2);
		if (!GetDimensionIx(group, var.getDim(0), dim1value, &ix[0]) || !GetDimensionIx(group, var.getDim(1), dim2value, &ix[1])) {
			return false;
		} else {
			var.getVar(ix, value);
			return true;
		}
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%s]: %s", group_name.c_str(), variable_name.c_str(), dim1value.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValue(const std::string& group_name, const std::string& variable_name, const std::string& dim1value, const std::string& dim2value, const std::string& dim3value, Real* value) const
{
	ASSERT(value);

	*value = -std::numeric_limits<Real>::quiet_NaN();

	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);

		std::vector<size_t> ix(3);
		if (!GetDimensionIx(group, var.getDim(0), dim1value, &ix[0]) || !GetDimensionIx(group, var.getDim(1), dim2value, &ix[1]) || !GetDimensionIx(group, var.getDim(2), dim3value, &ix[2])) {
			return false;
		} else {
			var.getVar(ix, value);
			return true;
		}
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%s]: %s", group_name.c_str(), variable_name.c_str(), dim1value.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValues(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t count, VectorReal& values) const
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(1);
		ix[0] = dim1ix;
		std::vector<size_t> countv(1);
		countv[0] = count;
		values.resize(count);
		var.getVar(ix, countv, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%u]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValuesDim1(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, size_t count, VectorReal& values) const
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(2);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		std::vector<size_t> countv(2);
		countv[0] = count;
		countv[1] = 1;
		values.resize(count);
		var.getVar(ix, countv, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%u,%u]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, dim2ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValuesDim2(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, size_t count, VectorReal& values) const
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(2);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		std::vector<size_t> countv(2);
		countv[0] = 1;
		countv[1] = count;
		values.resize(count);
		var.getVar(ix, countv, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%u,%u]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, dim2ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValuesDim1(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, size_t dim3ix, size_t count, VectorReal& values) const
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(3);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		ix[2] = dim3ix;
		std::vector<size_t> countv(3);
		countv[0] = count;
		countv[1] = 1;
		countv[2] = 1;
		values.resize(count);
		var.getVar(ix, countv, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%u,%u,%u]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, dim2ix, dim3ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValuesDim2(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, size_t dim3ix, size_t count, VectorReal& values) const
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(3);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		ix[2] = dim3ix;
		std::vector<size_t> countv(3);
		countv[0] = 1;
		countv[1] = count;
		countv[2] = 1;
		values.resize(count);
		var.getVar(ix, countv, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%u,%u,%u]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, dim2ix, dim3ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValuesDim3(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, size_t dim3ix, size_t count, VectorReal& values) const
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(3);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		ix[2] = dim3ix;
		std::vector<size_t> countv(3);
		countv[0] = 1;
		countv[1] = 1;
		countv[2] = count;
		values.resize(count);
		var.getVar(ix, countv, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving value for %s/%s[%u,%u,%u]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, dim2ix, dim3ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::GetValues(const std::string& group_name, const std::string& variable_name, std::vector<size_t> ix, std::vector<size_t> count, Real* values) const
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		var.putVar(ix, count, values);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s: %s", group_name.c_str(), variable_name.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::VariableExists(const std::string& group_name, const std::string& variable_name) const
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		return !var.isNull();
	} catch(netCDF::exceptions::NcException&) {
		return false;
	}
}

bool NetCDFDataFile::GetDimensionCount(const std::string& group_name, const std::string& variable_name, size_t* count) const
{
	ASSERT(count);
	*count = 0;
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		*count = (size_t)(var.getDimCount());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving number of dimensions for %s/%s: %s", group_name.c_str(), variable_name.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::GetDimensionName(const std::string& group_name, const std::string& variable_name, size_t dimix, std::string& dimname) const
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		netCDF::NcDim dim = var.getDim((int)dimix);
		dimname = dim.getName();
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving dimension size for %s/%s: %s", group_name.c_str(), dimname.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::GetDimensionSize(const std::string& group_name, const std::string& dimname, size_t* size) const
{
	ASSERT(size);
	*size = 0;
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcDim dim = group.getDim(dimname);
		*size = dim.getSize();
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving dimension size for %s/%s: %s", group_name.c_str(), dimname.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::GetDimensionIx(const std::string& group_name, const std::string& dimname, const std::string& dimvalue, size_t* ix) const
{
	ASSERT(ix);
	*ix = std::numeric_limits<size_t>::max();
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcDim dim = group.getDim(dimname);
		return GetDimensionIx(group, dim, dimvalue, ix);
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving index for %s/%s[%s]: %s", group_name.c_str(), dimname.c_str(), dimvalue.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::GetAttribute(const std::string& group_name, const std::string& attribute_name, Real& value) const
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcGroupAtt att = group.getAtt(attribute_name);
		att.getValues(&value);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error retrieving attribute %s: %s", attribute_name.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::CreateGroup(const std::string& group_name)
{
	if (!File) {
		LOGERROR("File not open");
		return false;
	}

	try {
		std::vector<std::string> hierarchy;
		bcm::tokenize(group_name, hierarchy, "/");

		netCDF::NcGroup group = *File;

		for (size_t i = 0; i < hierarchy.size(); i++) {
			// Workaround for bug in netcdf c++ library?
			int groupcount = group.getGroupCount();
			if (groupcount > 0) {
				netCDF::NcGroup child = group.getGroup(hierarchy[i]);
				if (child.isNull()) {
					group = group.addGroup(hierarchy[i]);
				} else {
					group = child;
				}
			} else {
				group = group.addGroup(hierarchy[i]);
			}
		}
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error creating new group %s: %s", group_name.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::CreateDimension(const std::string& group_name, const std::string& dimname, const std::vector<std::string>& dimvalues)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		if (dimvalues.empty()) {
			netCDF::NcDim dim = group.addDim(dimname);
			netCDF::NcVar dimvar = group.addVar(dim.getName(), netCDF::NcType::nc_STRING, dim);
		} else {
			netCDF::NcDim dim = group.addDim(dimname, dimvalues.size());
			CreateDimensionalVariable(group, dim, dimvalues);
		}
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error creating new dimension %s/%s: %s", group_name.c_str(), dimname.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::CreateDimension(const std::string& group_name, const std::string& dimname, const std::vector<unsigned int>& dimvalues)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		if (dimvalues.empty()) {
			netCDF::NcDim dim = group.addDim(dimname);
			netCDF::NcVar dimvar = group.addVar(dim.getName(), netCDF::NcType::nc_UINT, dim);
		} else {
			netCDF::NcDim dim = group.addDim(dimname, dimvalues.size());
			CreateDimensionalVariable(group, dim, dimvalues);
		}
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error creating new dimension %s/%s: %s", group_name.c_str(), dimname.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::CreateDimension(const std::string& group_name, const std::string& dimname, const std::vector<Real>& dimvalues)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		if (dimvalues.empty()) {
			netCDF::NcDim dim = group.addDim(dimname);
			netCDF::NcVar dimvar = group.addVar(dim.getName(), netCDF::NcType::nc_FLOAT, dim);
		} else {
			netCDF::NcDim dim = group.addDim(dimname, dimvalues.size());
			CreateDimensionalVariable(group, dim, dimvalues);
		}
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error creating new dimension %s/%s: %s", group_name.c_str(), dimname.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::CreateDimension(const std::string& group_name, const std::string& dimname, const VectorReal& dimvalues)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		if (dimvalues.size() == 0) {
			netCDF::NcDim dim = group.addDim(dimname);
			netCDF::NcVar dimvar = group.addVar(dim.getName(), netCDF::NcType::nc_FLOAT, dim);
		} else {
			netCDF::NcDim dim = group.addDim(dimname, dimvalues.size());
			CreateDimensionalVariable(group, dim, dimvalues);
		}
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error creating new dimension %s/%s: %s", group_name.c_str(), dimname.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::CreateVariable(const std::string& group_name, const std::string& variable_name, const std::string& dim1name)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.addVar(variable_name, netCDF::NcType::nc_FLOAT, group.getDim(dim1name, netCDF::NcGroup::ParentsAndCurrent));
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error creating new variable %s/%s: %s", group_name.c_str(), variable_name.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::CreateVariable(const std::string& group_name, const std::string& variable_name, const std::string& dim1name, const std::string& dim2name)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		std::vector<netCDF::NcDim> dims;
		dims.push_back(group.getDim(dim1name, netCDF::NcGroup::ParentsAndCurrent));
		dims.push_back(group.getDim(dim2name, netCDF::NcGroup::ParentsAndCurrent));
		netCDF::NcVar var = group.addVar(variable_name, netCDF::NcType::nc_FLOAT, dims);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error creating new variable %s/%s: %s", group_name.c_str(), variable_name.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::CreateVariable(const std::string& group_name, const std::string& variable_name, const std::string& dim1name, const std::string& dim2name, const std::string& dim3name)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		std::vector<netCDF::NcDim> dims;
		dims.push_back(group.getDim(dim1name, netCDF::NcGroup::ParentsAndCurrent));
		dims.push_back(group.getDim(dim2name, netCDF::NcGroup::ParentsAndCurrent));
		dims.push_back(group.getDim(dim3name, netCDF::NcGroup::ParentsAndCurrent));
		netCDF::NcVar var = group.addVar(variable_name, netCDF::NcType::nc_FLOAT, dims);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error creating new variable %s/%s: %s", group_name.c_str(), variable_name.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::CreateVariable(const std::string& group_name, const std::string& variable_name, const std::string& dim1name, const std::string& dim2name, const std::string& dim3name, const std::string& dim4name)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		std::vector<netCDF::NcDim> dims;
		dims.push_back(group.getDim(dim1name, netCDF::NcGroup::ParentsAndCurrent));
		dims.push_back(group.getDim(dim2name, netCDF::NcGroup::ParentsAndCurrent));
		dims.push_back(group.getDim(dim3name, netCDF::NcGroup::ParentsAndCurrent));
		dims.push_back(group.getDim(dim4name, netCDF::NcGroup::ParentsAndCurrent));
		netCDF::NcVar var = group.addVar(variable_name, netCDF::NcType::nc_FLOAT, dims);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error creating new variable %s/%s: %s", group_name.c_str(), variable_name.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValue(const std::string& group_name, const std::string& variable_name, const std::string& dim1value, Real value)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(1);
		if (!GetDimensionIx(group, var.getDim(0), dim1value, &ix[0])) {
			return false;
		} else {
			var.putVar(ix, (float)value);
			return true;
		}
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%s]: %s", group_name.c_str(), variable_name.c_str(), dim1value.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValue(const std::string& group_name, const std::string& variable_name, const std::string& dim1value, const std::string& dim2value, Real value)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(2);
		if (!GetDimensionIx(group, var.getDim(0), dim1value, &ix[0]) || !GetDimensionIx(group, var.getDim(1), dim2value, &ix[1])) {
			return false;
		} else {
			var.putVar(ix, (float)value);
			return true;
		}
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%s]: %s", group_name.c_str(), variable_name.c_str(), dim1value.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValue(const std::string& group_name, const std::string& variable_name, const std::string& dim1value, size_t dim2ix, Real value)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(2);
		ix[1] = dim2ix;
		if (!GetDimensionIx(group, var.getDim(0), dim1value, &ix[0])) {
			return false;
		} else {
			var.putVar(ix, (float)value);
			return true;
		}
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%s,%u]: %s", group_name.c_str(), variable_name.c_str(), dim1value.c_str(), dim2ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValue(const std::string& group_name, const std::string& variable_name, size_t dim1ix, unsigned int value)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(1);
		ix[0] = dim1ix;
		var.putVar(ix, value);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%u,]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValue(const std::string& group_name, const std::string& variable_name, size_t dim1ix, Real value)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(1);
		ix[0] = dim1ix;
		var.putVar(ix, (float)value);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%u,]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValue(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, Real value)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(2);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		var.putVar(ix, (float)value);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%u,%u]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, dim2ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValue(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, size_t dim3ix, Real value)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(3);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		ix[2] = dim3ix;
		var.putVar(ix, (float)value);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%u,%u,%u]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, dim2ix, dim3ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValue(const std::string& group_name, const std::string& variable_name, size_t dim1ix, const std::string& dim2value, size_t dim3ix, Real value)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(3);
		ix[0] = dim1ix;
		ix[2] = dim3ix;
		if (!GetDimensionIx(group, var.getDim(1), dim2value, &ix[1])) {
			return false;
		} else {
			var.putVar(ix, (float)value);
			return true;
		}
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%u,%s,%u]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, dim2value.c_str(), dim3ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValues(const std::string& group_name, const std::string& variable_name, size_t dim1ix, const VectorReal& values)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(1);
		ix[0] = dim1ix;
		std::vector<size_t> count(1);
		count[0] = values.size();
		var.putVar(ix, count, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%u,]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValuesDim1(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, const VectorReal& values)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(2);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		std::vector<size_t> count(2);
		count[0] = values.size();
		count[1] = 1;
		var.putVar(ix, count, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%u,]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValuesDim2(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, const VectorReal& values)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(2);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		std::vector<size_t> count(2);
		count[0] = 1;
		count[1] = values.size();
		var.putVar(ix, count, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%u,]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValuesDim1(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, size_t dim3ix, const VectorReal& values)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(3);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		ix[2] = dim3ix;
		std::vector<size_t> count(3);
		count[0] = values.size();
		count[1] = 1;
		count[2] = 1;
		var.putVar(ix, count, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%u,]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValuesDim2(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, size_t dim3ix, const VectorReal& values)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(3);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		ix[2] = dim3ix;
		std::vector<size_t> count(3);
		count[0] = 1;
		count[1] = values.size();
		count[2] = 1;
		var.putVar(ix, count, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%u,]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValuesDim3(const std::string& group_name, const std::string& variable_name, size_t dim1ix, size_t dim2ix, size_t dim3ix, const VectorReal& values)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		
		std::vector<size_t> ix(3);
		ix[0] = dim1ix;
		ix[1] = dim2ix;
		ix[2] = dim3ix;
		std::vector<size_t> count(3);
		count[0] = 1;
		count[1] = 1;
		count[2] = values.size();
		var.putVar(ix, count, values.data());
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s[%u,]: %s", group_name.c_str(), variable_name.c_str(), dim1ix, e.what());
		return false;
	}
}

bool NetCDFDataFile::PutValues(const std::string& group_name, const std::string& variable_name, const std::vector<size_t>& ix, const std::vector<size_t>& count, const Real* values)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		netCDF::NcVar var = group.getVar(variable_name);
		var.putVar(ix, count, values);
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error setting value for %s/%s: %s", group_name.c_str(), variable_name.c_str(), e.what());
		return false;
	}
}

bool NetCDFDataFile::GetSubgroups(const std::string& group_name, std::vector<std::string>& groups)
{
	try {
		netCDF::NcGroup group = GetGroup(group_name);
		std::multimap<std::string, netCDF::NcGroup> ncdf_groups = group.getGroups();
		for (std::multimap<std::string, netCDF::NcGroup>::iterator gi = ncdf_groups.begin(); gi != ncdf_groups.end(); ++gi) {
			groups.push_back(gi->first);
		}
		return true;
	} catch(netCDF::exceptions::NcException& e) {
		LOGERROR("Error getting subgroups for %s: %s", group_name.c_str(), e.what());
		return false;
	}
}

netCDF::NcGroup NetCDFDataFile::GetGroup(const std::string& group_name) const
{
	if (!File) {
		LOGERROR("File not open");
		return netCDF::NcGroup();
	}

	std::vector<std::string> hierarchy;
	bcm::tokenize(group_name, hierarchy, "/");

	netCDF::NcGroup group = *File;
	for (size_t i = 0; i < hierarchy.size(); i++) {
		group = group.getGroup(hierarchy[i]);
	}

	return group;
}

void NetCDFDataFile::CreateDimensionalVariable(netCDF::NcGroup group, netCDF::NcDim dim, const std::vector<std::string>& dimvalues)
{
	netCDF::NcVar dimvar = group.addVar(dim.getName(), netCDF::NcType::nc_STRING, dim);
	std::vector<const char*> dimvalues_cptr;
	dimvalues_cptr.resize(dimvalues.size());
	for (size_t i = 0; i < dimvalues.size(); i++) {
		dimvalues_cptr[i] = dimvalues[i].c_str();
	}
	dimvar.putVar(&dimvalues_cptr[0]);
}

void NetCDFDataFile::CreateDimensionalVariable(netCDF::NcGroup group, netCDF::NcDim dim, const std::vector<unsigned int>& dimvalues)
{
	netCDF::NcVar dimvar = group.addVar(dim.getName(), netCDF::NcType::nc_UINT, dim);
	dimvar.putVar(&dimvalues[0]);
}

void NetCDFDataFile::CreateDimensionalVariable(netCDF::NcGroup group, netCDF::NcDim dim, const std::vector<Real>& dimvalues)
{
	netCDF::NcVar dimvar = group.addVar(dim.getName(), netCDF::NcType::nc_FLOAT, dim);
	std::vector<float> dimvalues_float;
	dimvalues_float.resize(dimvalues.size());
	for (size_t i = 0; i < dimvalues.size(); i++) {
		dimvalues_float[i] = (float)dimvalues[i];
	}
	dimvar.putVar(&dimvalues_float[0]);
}

void NetCDFDataFile::CreateDimensionalVariable(netCDF::NcGroup group, netCDF::NcDim dim, const VectorReal& dimvalues)
{
	ASSERT(dimvalues.size() >= 0);
	netCDF::NcVar dimvar = group.addVar(dim.getName(), netCDF::NcType::nc_FLOAT, dim);
	std::vector<float> dimvalues_float;
	dimvalues_float.resize((size_t)dimvalues.size());
	for (size_t i = 0; i < (size_t)dimvalues.size(); i++) {
		dimvalues_float[i] = (float)dimvalues[i];
	}
	dimvar.putVar(&dimvalues_float[0]);
}

bool NetCDFDataFile::GetDimensionIx(netCDF::NcGroup group, netCDF::NcDim dim, const std::string& dimvalue, size_t* ix) const
{
	netCDF::NcVar dimvar = group.getVar(dim.getName());
	if (dimvar.isNull()) {
		// Try parent group
		while (!group.getParentGroup().isNull()) {
			dimvar = group.getParentGroup().getVar(dim.getName());
			if (!dimvar.isNull()) {
				break;
			}
		}
	}
	
	*ix = dim.getSize();
	std::vector<size_t> index(1);
	for (size_t i = 0; i < dim.getSize(); i++) {
		char* name;
		index[0] = i;
		dimvar.getVar(index, &name);
		if (strcmp(name, dimvalue.c_str()) == 0) {
			*ix = i;
			free(name);
			break;
		} else {
			free(name);
		}
	}

	if (*ix == dim.getSize()) {
		LOGERROR("Error retrieving value for %s/%s[%s]: could not find dimension value", group.getName().c_str(), dim.getName().c_str(), dimvalue.c_str());
		return false;
	} else {
		return true;
	}
}

}

#endif
