#pragma once

#include <boost/math/special_functions/log1p.hpp>

namespace bcm {

class RNG
{
public:
	virtual ~RNG();

	virtual void Seed(unsigned long seed) = 0;

	virtual unsigned long	GetInt() = 0;										//< Generates uniform random integer in interval [0,0xffffffff]
	virtual Real			GetReal() = 0;										//< Generates uniform random number in interval [0,1)
	inline	Real			GetUniform(const Real lower, const Real upper);		//< Generates uniform random number in interval [lower,upper)
	inline	unsigned long	GetUniformInt(const unsigned long upper);			//< Generates uniform random integer in interval [0,upper)
	inline	Real			GetExponential(const Real mu);
			Real			GetGamma(const Real k, const Real theta);
			Real			GetNormal(const Real mu, const Real sigma);
	inline  Real			GetBernoulli(const Real p);
	inline	Real			GetBeta(const Real a, const Real b);
	inline	Real			GetCauchy(const Real scale);
	inline	Real			GetHalfCauchy(const Real scale);

	static unsigned long GetTimeBasedSeed();

private:
};

inline Real RNG::GetUniform(const Real lower, const Real upper)
{
	ASSERT(lower <= upper);
	return lower + GetReal() * (upper - lower);
}

inline unsigned long RNG::GetUniformInt(const unsigned long upper)
{
	if (upper == 0) {
		return 0;
	}

	unsigned long scale = 0xfffffffful / upper;
	unsigned long k;
	do {
		k = GetInt() / scale;
	} while (k >= upper);
	return k;
}

inline Real RNG::GetExponential(const Real mu)
{
	const Real u = GetReal();
	return -mu * boost::math::log1p(-u);
}

inline Real RNG::GetBernoulli(const Real p)
{
	if (GetReal() < p) {
		return 1.0;
	} else {
		return 0.0;
	}
}

inline Real RNG::GetBeta(const Real a, const Real b)
{
	Real x1 = GetGamma(a, 1.0);
	Real x2 = GetGamma(b, 1.0);
	return x1 / (x1 + x2);
}

inline Real RNG::GetCauchy(const Real scale)
{
	Real x1 = GetReal();
	return scale * tan((Real)M_PI * (x1 - (Real)0.5));
}

inline Real RNG::GetHalfCauchy(const Real scale)
{
	return fabs(GetCauchy(scale));
}

}
